#!/usr/bin/env python

from PySide import QtCore, QtGui
from sherpassexception import SherPassException
import ui_configdialog
from keydialog import KeyDialog
from generatekeydialog import GenerateKeyDialog
from exceptionwarning import ExceptionWarning as EW
import sys
import os
import sharedpass
import platform
import randomstring
import logo

def _findExeInPath(fileName):

    path = os.environ.get("PATH")
    if not path:
        return False

    isWin = True if platform.system() == "Windows" else False

    paths = path.split(os.pathsep)
    for p in paths:
        fullFileName = os.path.join(p, fileName)
        if os.path.isfile(fullFileName):
            if isWin:
                return True
            if os.access(fullFileName, os.X_OK):
                return True
    
    return False

class ConfigDialog(QtGui.QDialog):
    def __init__(self, confDir, confData, parent):

        super(ConfigDialog, self).__init__(parent)

        self.ui = ui_configdialog.Ui_ConfigDialog()
        self.ui.setupUi(self)

        self.ui.cancelButton.clicked.connect(self.reject)
        self.ui.saveButton.clicked.connect(EW(self.onSave))
        self.ui.generateButton.clicked.connect(EW(self.onGenerate))
        self.ui.importButton.clicked.connect(EW(self.onImport))
        self.ui.browseKeyButton.clicked.connect(EW(self.onBrowsePubKeyDir))
        self.ui.browsePassButton.clicked.connect(EW(self.onBrowsePassDir))
        self.ui.tcpGenerateButton.clicked.connect(EW(self.onGenerateTCPCodes))
        self.ui.iconLabel.setPixmap(logo.getPixmap())

        self.ui.configDirEdit.setText(confDir)

        if confData:
            if "PubKeyDir" in confData: 
                self.ui.pubKeyDirEdit.setText(confData["PubKeyDir"])

            if "PassDir" in confData:
                self.ui.passDirEdit.setText(confData["PassDir"])

        if confData and "SshCommand" in confData:
            self.ui.sshCommandEdit.setText(confData["SshCommand"])
        else:
            # This is the fallback command
            cmd = "xterm -T \"%u@%h\" -e \"sshpass -e ssh -p %P %u@%h ; sleep 5\""

            if platform.system() == "Windows":
                cmd = "putty.exe -pw %p -l %u -P %P %h"
            elif platform.system() == "Darwin":
                cmd = "qterminal -e sshpass -e ssh -p %P %u@%h"
            else: # Assume it's some linux
                if _findExeInPath("gnome-terminal"):
                    cmd = 'gnome-terminal -t "%u@%h" -e "sshpass -e ssh -p %P %u@%h"'

            self.ui.sshCommandEdit.setText(cmd)

        if confData and "SshPassEnv" in confData:
            if confData["SshPassEnv"]: # just to make sure the placeholder can be shown
                self.ui.passEnvVarEdit.setText(confData["SshPassEnv"])
        else:
            if platform.system() != "Windows": # For linux we're using SSHPASS by default
                self.ui.passEnvVarEdit.setText("SSHPASS")

        if confData:
            if "TcpPort" in confData and confData["TcpPort"] > 0:
                self.ui.tcpEnableCheckBox.setCheckState(QtCore.Qt.Checked)
                if confData["TcpPort"] < 65536:
                    self.ui.tcpPortSpinBox.setValue(confData["TcpPort"])
                else:
                    self.ui.tcpPortSpinBox.setValue(49374)

            if "AccessUrl" in confData:
                self.ui.accessUrlEdit.setText(confData["AccessUrl"])
            if "TcpPassPhrase" in confData:
                self.ui.aesKeyEdit.setText(confData["TcpPassPhrase"])

        self.pubKey = None
        self.privKey = None
        self.confData = None # This is the outgoing data

    def onBrowse(self, widget, mode):
        dlg = QtGui.QFileDialog()
        if mode:
            dlg.setFileMode(mode)
        dlg.setDirectory(widget.text())
        if dlg.exec_():
            f = dlg.selectedFiles()
            if f:
                widget.setText(f[0])

    def onBrowsePassDir(self):
        self.onBrowse(self.ui.passDirEdit, QtGui.QFileDialog.Directory)

    def onBrowsePubKeyDir(self):
        self.onBrowse(self.ui.pubKeyDirEdit, QtGui.QFileDialog.Directory)

    def getConfigurationData(self):
        return self.confData

    def onSave(self):
        self.confData = { "PubKeyDir": self.ui.pubKeyDirEdit.text(),
                          "PassDir": self.ui.passDirEdit.text(),
                          "PrivKey": self.privKey,
                          "PubKey": self.pubKey,
                          "SshCommand": self.ui.sshCommandEdit.text(),
                          "TcpPort": self.ui.tcpPortSpinBox.value() if self.ui.tcpEnableCheckBox.checkState() == QtCore.Qt.Checked else 0,
                          "AccessUrl": self.ui.accessUrlEdit.text(),
                          "TcpPassPhrase": self.ui.aesKeyEdit.text(),
                          "SshPassEnv": self.ui.passEnvVarEdit.text() }
        self.accept()

    def onGenerate(self):
        dlg = GenerateKeyDialog(self)
        if dlg.exec_() != QtGui.QDialog.Accepted:
            return

        (self.privKey, self.pubKey) = dlg.getKeys()

    def onImport(self):
        dlg = KeyDialog(None, self)
        if dlg.exec_() != QtGui.QDialog.Accepted:
            return
        
        privKey = dlg.getKey()
        try:
            sp = sharedpass.SharedPass(privKey, None, None, sync = False)
            self.pubKey = sp.exportPublicKey()
            self.privKey = sp.exportPrivateKey()

            #print self.pubKey
            #print self.privKey

            QtGui.QMessageBox.information(self, "Done", "Private key imported and associated public key generated")
        except SherPassException as e:
            QtGui.QMessageBox.warning(self, "Invalid key", "Unable to use the imported private key: " + str(e))

    def onGenerateTCPCodes(self):
        self.ui.aesKeyEdit.setText(randomstring.getRandomString(16))
        self.ui.accessUrlEdit.setText(randomstring.getRandomString(16))

