#!/usr/bin/env python

from sherpassexception import SherPassException
import os
import sys
import gnupg
import randomstring
import json
import copy
import sptempdir
import gpgdir

class SharedPass(object):

    rndStringLength = 16

    def __init__(self, privateKey, pubKeyDir, passDir, sync = True, passPhrase = None):

        self.gpgTempDir = sptempdir.getTempDir()
        print "Created directory %s" % self.gpgTempDir

        gpg = gnupg.GPG(gpgbinary=gpgdir.gpg, gnupghome=self.gpgTempDir)
        #gpg = gnupg.GPG(gnupghome="./gpgdir/",verbose=True)

        gpg.import_keys(privateKey)
        
        privKeys = gpg.list_keys(True)
        if len(privKeys) != 1:
            raise SherPassException("There should be exactly one private key in the GPG keystore")

        pubKeys = gpg.list_keys(False)
        if len(pubKeys) != 1:
            raise SherPassException("There should not be any extra public keys present during initialization (private key file contains other public key data?)")

        self.ownFingerPrint = privKeys[0]["fingerprint"]
        
        self.gpg = gpg
        self.pubKeyDir = pubKeyDir
        self.passDir = passDir
        self.passInfo = [ ]
        self.passPhrase = passPhrase

        if sync:
            self.syncPublicKeys()
            self.syncPassInfo()

    def __del__(self): # Must delete the temporary directory again
        files = os.listdir(self.gpgTempDir)
        for f in files:
            fullName = os.path.join(self.gpgTempDir, f)
            os.unlink(fullName)
            print "Removed %s" % fullName

        os.rmdir(self.gpgTempDir)
        print "Removed %s" % self.gpgTempDir

    def getOwnFingerprint(self):
        return self.ownFingerPrint

    def setPassPhrase(self, passPhrase):
        self.passPhrase = passPhrase

    def isPubKeyFileNameValid(self, fileName):
        if fileName[0] == ".":
            return False
        suffix = ".pubkey"
        if fileName[-len(suffix):].lower() != suffix:
            return False
        return True

    def isPassFileNameValid(self, fileName):
        if fileName[0] == ".":
            return False
        suffix = ".pass"
        if fileName[-len(suffix):].lower() != suffix:
            return False
        return True

    def syncPublicKeys(self):

        gpg = self.gpg

        oldFingerprints = set()
        pubKeys = gpg.list_keys()
        for p in pubKeys:
            fp = p["fingerprint"]
            if fp != self.ownFingerPrint:
                oldFingerprints.add(fp)
        
        pubKeyFiles = os.listdir(self.pubKeyDir)
        importedFingerPrints = set()
        for fileName in pubKeyFiles:
            if not self.isPubKeyFileNameValid(fileName):
                raise SherPassException("Invalid public key filename: " + fileName)

            fullFileName = os.path.join(self.pubKeyDir,fileName)
            with open(fullFileName,"rt") as f:
                keyData = f.read()
                f.close()

            pubKeys = gpg.import_keys(keyData)
            for fp in pubKeys.fingerprints:
                importedFingerPrints.add(fp)

        privKeys = gpg.list_keys(True)
        if len(privKeys) != 1: # Remove the other private keys again
            delKeys = ""
            for k in privKeys:
                fp = k["fingerprint"]
                if fp != self.ownFingerPrint:
                    delKeys += " " + fp
                    gpg.delete_keys([fp],True) # Private key
                    gpg.delete_keys([fp])      # Public key (if it exists)

            raise SherPassException("Imported other private keys (deleted them again):" + delKeys)

        if not self.ownFingerPrint in importedFingerPrints:
            raise SherPassException("Own public key is not among the imported keys")

        # Remove the fingerprints from our keystore that are no longer in the
        # key directory

        for fp in oldFingerprints:
            if not fp in importedFingerPrints: # Own key is also imported
                gpg.delete_keys([fp])

    def listPublicKeys(self):
        gpg = self.gpg
        pubKeys = gpg.list_keys()
        keyList = [ ]
        for k in pubKeys:
            keyList.append( (k["fingerprint"], k["uids"][0]) )

        return keyList
                
    def exportPublicKey(self):
        gpg = self.gpg
        s = str(gpg.export_keys([self.ownFingerPrint]))
        if not s:
            raise SherPassException("Could not find public key with own fingerprint in keystore")

        return s

    def exportPrivateKey(self):
        gpg = self.gpg
        s = str(gpg.export_keys([self.ownFingerPrint], True))
        if not s:
            raise SherPassException("Could not find private key with own fingerprint in keystore")

        return s

    def encrypt(self, s, fingerPrints):

        #print "SharedPass.encrypt: ", fingerPrints
        #print "data:", s

        s = randomstring.getRandomString(self.rndStringLength) + s

        gpg = self.gpg
        pubKeys = gpg.list_keys()
        fpList = [ ]

        if not fingerPrints: # Means all known fingerprints
            for k in pubKeys:
                fpList.append(k["fingerprint"])
        else:
            knownPrints = set([ k["fingerprint"] for k in pubKeys ])
            for k in fingerPrints:
                if k in knownPrints:
                    fpList.append(k)
                else:
                    raise SherPassException("An unknown fingerprint was specified")

            if not self.ownFingerPrint in fpList:
                raise SherPassException("Own fingerprint not specified, disallowing encoding data that we can't decode ourselves")

        # TODO: Not sure how safe this 'always trust' is in this scenario. Didn't need it when using a
        #       gpg keyring from the start, but do need it when just starting from scratch and importing
        #       a private key

        #print "fpList", fpList

        encData = str(gpg.encrypt(s, fpList, always_trust=True)) 
        if not encData:
            raise SherPassException("Error encrypting data (invalid key fingerprint?)")

        return encData

    def checkCurrentPassphrase(self):

        msg = randomstring.getRandomString(32)
        gpg = self.gpg
        
        s = str(gpg.encrypt(msg, [ self.getOwnFingerprint() ], always_trust=True))
        if not self.passPhrase:
            decData = str(gpg.decrypt(s))
        else:
            decData = str(gpg.decrypt(s,passphrase=self.passPhrase))

        if not decData or decData != msg:
            raise SherPassException("Invalid passphrase")

    def decrypt(self, s):

        gpg = self.gpg
        if not self.passPhrase:
            decData = str(gpg.decrypt(s))
        else:
            decData = str(gpg.decrypt(s,passphrase=self.passPhrase))

        if not decData:
            raise SherPassException("Unable to decode specified data")

        if len(decData) < self.rndStringLength:
            raise SherPassException("Invalid decode result")

        return decData[self.rndStringLength:]

    def syncPassInfo(self):

        self.passInfo = [ ]
        undecodedFiles = [ ]
        
        passFiles = os.listdir(self.passDir)
        passFiles = [ fileName for fileName in passFiles if self.isPassFileNameValid(fileName) and len(fileName) == 16+1+4 ]
        
        for fileName in passFiles:
            #if not self.isPassFileNameValid(fileName):
            #    raise SherPassException("Invalid password filename: " + fileName)

            fullFileName = os.path.join(self.passDir,fileName)
            with open(fullFileName, "rt") as f:
                data = f.read()
                f.close()

            decData = None
            try:
                decData = self.decrypt(data)
            except SherPassException as e:
                undecodedFiles.append(fileName)

            if decData:
                obj = json.loads(decData)
                self.passInfo.append( (fileName, obj) )

        return (len(undecodedFiles) == 0, undecodedFiles, passFiles)

    def listPasswords(self):
        return copy.deepcopy(self.passInfo)

    def addPassword(self, obj, fingerPrints = None): # None means all known fingerprints
        s = json.dumps(obj)
        enc = self.encrypt(s, fingerPrints)
        fileName = randomstring.getRandomString(16) + ".pass"
        fullFileName = os.path.join(self.passDir,fileName)

        if os.path.exists(fullFileName):
            raise SherPassException("Randomly generated filename already seems to exist: " + fileName)

        with open(fullFileName, "wt") as f:
            f.write(enc)
            f.close()

        decObj = json.loads(s)
        newEntry = (fileName, decObj)
        self.passInfo.append(newEntry)

        return copy.deepcopy(newEntry)

    def updatePassword(self, fileName, obj, fingerPrints = None): # None just means all loaded prints

        entryIdx = -1
        for idx in range(len(self.passInfo)):
            if self.passInfo[idx][0] == fileName:
                entryIdx = idx
                break

        if entryIdx < 0:
            raise SherPassException("No entry with the specified filename found: " + fileName)

        s = json.dumps(obj)
        enc = self.encrypt(s, fingerPrints)
        fullFileName = os.path.join(self.passDir,fileName)

        if not os.path.exists(fullFileName):
            raise SherPassException("Specified filename does not yet exist: " + fileName)

        with open(fullFileName, "wt") as f:
            f.write(enc)
            f.close()

        decObj = json.loads(s)
        newEntry = (fileName, decObj)
        entries = self.passInfo

        self.passInfo = entries[:entryIdx] + [ newEntry ] + entries[entryIdx+1:]
        return copy.deepcopy(newEntry)

    def removePassword(self, fileName):

        entryIdx = -1
        for idx in range(len(self.passInfo)):
            if self.passInfo[idx][0] == fileName:
                entryIdx = idx
                break

        if entryIdx < 0:
            raise SherPassException("No entry with the specified filename found: " + fileName)

        fullFileName = os.path.join(self.passDir, fileName)
        os.unlink(fullFileName)
        entries = self.passInfo

        self.passInfo = entries[:entryIdx] + entries[entryIdx+1:]

