#!/usr/bin/env python

import subprocessmod
from PySide import QtCore, QtGui, QtNetwork
from sherpassexception import SherPassException
import ui_sherpass
from entrydialog import EntryDialog
from configdialog import ConfigDialog
from fingerprintdialog import FingerprintDialog
from passphrasedialog import PassphraseDialog
from keydialog import KeyDialog
from newpassworddialog import NewPasswordDialog
from changesshpasswordsdialog import ChangeSSHPasswordsDialog
from logdialog import LogDialog
from exceptionwarning import ExceptionWarning as EW
import decodingpasswordsdialog
import sys
import os
import json
import sharedpass
import subprocess
import shlex
import pprint
import sptempdir
import time
import aescrypt
import logo
import copy
import pprint
import randomstring
import licenses
import portalocker
import popupmenu
import platform



#TODO: remove this! for debugging
#def EW(f):
#    return f

lockFile = None

# We'll just open a file for writing as a lock file, and keep it
# open until the process exits

class LockFile(object):
    def __init__(self, fileName):
        self.fileDesc = None
        self.fileDesc = open(fileName, "wt")
        portalocker.lock(self.fileDesc, portalocker.LOCK_EX|portalocker.LOCK_NB)

        self.fileDesc.write(str(os.getpid()))
        self.fileDesc.flush()

    def __del__(self):
        try:
            if self.fileDesc:
                self.fileDesc.close()
                self.fileDesc = None
        except:
            pass

class SherPass(QtGui.QMainWindow):

    def __init__(self, configDir, parent = None):
        super(SherPass, self).__init__(parent)

        self.setWindowIcon(logo.getLogo())

        self.configDir = configDir
        self.configDir = os.path.abspath(self.configDir)

        self.passPhrase = ""
        self.confData = None
        self.sharedPass = None

        self.tcpServer = None
        self.connections = [ ]
        self.badLocationCount = 0
        self.connectionLog = ""

        self.ui = ui_sherpass.Ui_SherPass()
        self.ui.setupUi(self)

        self.scheduleAsync(self.onInitialTimeout)

        self.popupMenu = popupmenu.PopupMenu(None)

        connCheckTimer = QtCore.QTimer(self)
        connCheckTimer.timeout.connect(EW(self.onConnectionCheck, False))
        connCheckTimer.start(10000)

        self.ui.addEntryButton.clicked.connect(EW(self.onAddEntry)) 
        self.ui.changeEntryButton.clicked.connect(EW(self.onChangeEntry))
        self.ui.removeEntryButton.clicked.connect(EW(self.onRemoveEntry))
        self.ui.sshButton.clicked.connect(EW(self.onSSH))

        self.ui.passwordTable.cellDoubleClicked.connect(EW(self.onEntryDoubleClicked))
        self.ui.passwordTable.cellActivated.connect(EW(self.onCellActivated))
        self.ui.passwordTable.rightClickSignal.connect(EW(self.onCellRightClicked))

        self.ui.filterEdit.textChanged.connect(EW(self.ui.passwordTable.setFilter))
        self.ui.filterEdit.textEdited.connect(EW(self.ui.passwordTable.setFilter))
        self.ui.filterEdit.textChanged.connect(EW(self.onFilterChanged))
        self.ui.filterEdit.textEdited.connect(EW(self.onFilterChanged))
        self.ui.filterEdit.hide()
        self.ui.filterLabel.hide()

        self.quitting = True
        self.confirmQuit = False
        self.ui.actionConfigure.triggered.connect(EW(self.onConfigure))
        self.ui.actionQuit.triggered.connect(EW(self.onQuit))
        self.ui.actionShow_known_keys.triggered.connect(EW(self.onShowFingerprints))
        self.ui.actionRe_load_keys.triggered.connect(EW(self.onSyncPublicKeys))
        self.ui.actionSet_passphrase_re_load.triggered.connect(EW(self.onNewPassPhrase))
        self.ui.actionRe_load_password_entries.triggered.connect(EW(self.onSyncPassInfo))
        self.ui.actionRe_encode_all_known_entries.triggered.connect(EW(self.onRecodeAll))
        self.ui.actionExport_public_key.triggered.connect(EW(self.onExportPublicKey))
        self.ui.actionView_private_key.triggered.connect(EW(self.onViewPrivateKey))
        self.ui.actionView_public_key.triggered.connect(EW(self.onViewPublicKey))
        self.ui.actionChange_to_specific_password.triggered.connect(EW(self.onChangeToSpecificPassword))
        self.ui.actionChange_to_random_password.triggered.connect(EW(self.onChangeToRandomPassword))
        self.ui.actionTCP_Connection_log.triggered.connect(EW(self.onConnectionLog))
        self.ui.actionVersion.triggered.connect(EW(self.onVersion))
        self.ui.actionWeb_site.triggered.connect(EW(self.onWebSite))
        self.ui.actionExport_all_known_entries.triggered.connect(EW(self.onExportEntries))

        self.sysTrayIcon = QtGui.QSystemTrayIcon(logo.getLogo())
        self.sysTrayIcon.show()

        self.sysTrayIcon.activated.connect(EW(self.toggleShow))

        self.setFocusPolicy(QtCore.Qt.StrongFocus)

        settings = QtCore.QSettings()
        mainWinGeom = settings.value("mainWindowGeometry")
        mainWinState = settings.value("mainWindowState")
        if mainWinGeom:
            self.restoreGeometry(mainWinGeom)
        if mainWinState:
            self.restoreState(mainWinState)

    def closeAllConnections(self):

        if self.tcpServer:
            self.tcpServer.newConnection.disconnect(self.onNewConnection)
        self.tcpServer = None

        for (c, t) in self.connections:
            c.disconnected.disconnect(self.onDisconnected)
            c.readyRead.disconnect(self.onReadyRead)

        self.connections = [ ]

    def closeEvent(self, e):

        if not self.quitting:
            self.hide()
            e.ignore()
            return
        
        self.quitting = False

        if self.confirmQuit:
            if QtGui.QMessageBox.question(self, "Sure?", "Are you sure you want to exit?", QtGui.QMessageBox.Yes|QtGui.QMessageBox.No, QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
                e.ignore()
                return

        self.closeAllConnections()
        e.accept()

        settings = QtCore.QSettings()
        settings.setValue("mainWindowGeometry", self.saveGeometry())
        settings.setValue("mainWindowState", self.saveState())

    def getConfigDir(self):
        return self.configDir

    def getConfigFile(self):
        return os.path.join(self.getConfigDir(), "config.json")

    def loadConfig(self):
        confDir = self.getConfigDir()
        if not os.path.exists(confDir):
            raise SherPassException("Configuration base path '%s' does not exist" % confDir)
        if not os.path.isdir(confDir):
            raise SherPassException("Configuration base path '%s' exists, but is not a directory" % confDir)

        confFile = self.getConfigFile()
        f = None
        try:
            f = open(confFile, "rt")
            data = f.read()
        except Exception as e:
            raise SherPassException("Unable to read from config file '%s': " % str(e))
        finally:
            if f: f.close()

        try:
            confObj = json.loads(data)
        except Exception as e:
            raise SherPassException("Unable to parse config file: " + str(e))

        neededKeys = [ "PubKeyDir", "PassDir", "PrivateKey" ]
        optionalKeys = [ "TcpPort", "AccessUrl", "TcpPassPhrase", "SshCommand", "SshPassEnv", "Logo" ]
        newObj = { }
        for s in neededKeys:
            if not s in confObj:
                raise SherPassException("Config file entry '%s' was not found" % s)

            newObj[s] = confObj[s]

        for s in optionalKeys:
            if s in confObj:
                newObj[s] = confObj[s]

        return newObj

    def removeFileOrDir(self, path):
        if os.path.isdir(path):
            
            # First remove the contents, should all be regular files

            files = os.listdir(path)
            for fileName in files: # We're not expecting any further directories, and don't want to go recursive
                name = os.path.join(path, fileName)
                print "Removing", name
                os.unlink(name)
    
            # Then remove the directory itself
            print "Removing", path
            os.rmdir(path)
        else:
            print "Removing", path
            os.unlink(path)

    def getLockFileName(self):

        confDir = self.getConfigDir()
        lockName = os.path.join(confDir, "lockfile.lock")

        return lockName

    def couldBeRunning(self):
        global lockFile # hopefully the destructor will clean up on clean exit

        try:
            lockFile = LockFile(self.getLockFileName())
        except:
            return True

        return False

    def onInitialTimeout(self):

        try:
            tempDirBase = os.path.join(self.getConfigDir(), "gpgtempdirs")
            sptempdir.setTempDirBase(tempDirBase)
        except Exception as e:
            QtGui.QMessageBox.critical(self, "Fatal error setting temporary directory", "Fatal error setting temporary directory: " + str(e) + "\n\n\nExiting...")
            self.close()
            return
        
        if self.couldBeRunning():
            QtGui.QMessageBox.critical(self, "Already running", "An instance of SherPass using this config directory already seems to be running.\nDelete the file '%s' if not" % self.getLockFileName())
            self.close()
            return

        # If there's anything left in the temporary directory, we'll remove it
        try:
            files = os.listdir(tempDirBase)
            for fileName in files:
                name = os.path.join(tempDirBase, fileName)
                self.removeFileOrDir(name) # This is not recursive but should work for both a file and a directory (that doesn't contain any more directories)
        except Exception as e:
            QtGui.QMessageBox.warning(self, "Can't clear the temporary directory", "Error clearing the base temporary directory '%s'\n%s'" % (tempDirBase, str(e)))

        self.quitting = False
        self.confirmQuit = True
        self.restart()

    def scheduleAsync(self, function):
        t = QtCore.QTimer(self)
        t.timeout.connect(EW(function, False))
        t.setSingleShot(True)
        t.start(0)

    def setCustomLogo(self):
        try:
            confData = self.confData

            if confData and "Logo" in confData and confData["Logo"]:
                if confData["Logo"][0] == ":":
                    logoName = confData["Logo"][1:]
                    internalPixmaps = logo.getInternalPixmaps()
                    if logoName in internalPixmaps:
                        logo.setPixmap(internalPixmaps[logoName])
                    else:
                        raise SherPassException("Invalid internal logo name")
                
                else:
                    img = QtGui.QImage()
                    if not img.load(confData["Logo"]):
                        raise SherPassException("Unable to load specified custom logo from file %s" % confData["Logo"])
                    logo.setImage(img)

            else:
                logo.clearCustomLogo()

            self.setWindowIcon(logo.getLogo())
            self.sysTrayIcon.setIcon(logo.getLogo())

        except Exception as e:
            QtGui.QMessageBox.warning(self, "Can't set logo", "Unable to set custom logo:\n%s'" % str(e))

    def restart(self):
        
        self.closeAllConnections()

        while self.ui.passwordTable.rowCount():
            self.ui.passwordTable.removeRow(0)

        self.passPhrase = ""
        self.confData = None
        self.sharedPass = None

        try:
            confData = self.loadConfig()
            self.confData = confData
        except Exception as e:
            QtGui.QMessageBox.warning(self, "Can't initialize", "Can't get configuration data: %s" % str(e))
            self.scheduleAsync(self.onConfigure)
            return

        self.setCustomLogo()

        try:
            self.getPassPhrase() 
            
            with open(confData["PrivateKey"]) as f:
                keyData = f.read()

            sp = sharedpass.SharedPass(keyData, confData["PubKeyDir"], confData["PassDir"], passPhrase = self.passPhrase, sync = False)
            self.sharedPass = sp

            self.onSyncPublicKeys()
            self.onSyncPassInfo()

        except Exception as e:
            QtGui.QMessageBox.warning(self, "Can't initialize", "Unable to initialize application: %s" % str(e))
            return

        if "TcpPort" in confData and confData["TcpPort"] > 0:
            if "AccessUrl" in confData and len(confData["AccessUrl"]) >= 12 and "TcpPassPhrase" in confData and len(confData["TcpPassPhrase"]) >= 12:
                self.startTCPListener()
            else:
                QtGui.QMessageBox.warning(self, "TCP server not available", "The quick access TCP server is requested, but badly configured")

    def addOrUpdateEntry(self, row, name, passInfo, message = None):
        (gotWarning, filteredPrints) = self.filterFingerPrints(passInfo)
        if row < 0 or not name:
            newEntry = self.sharedPass.addPassword(passInfo, filteredPrints)
            self.appendEntryToTable(newEntry)
        else:
            newEntry = self.sharedPass.updatePassword(name, passInfo, filteredPrints)
            self.setRowEntry(row, newEntry)

        self.setEntryMessage(row, message)

        return gotWarning 

    def setEntryMessage(self, row, msg):
        if msg:
            self.ui.passwordTable.setItem(row, 4, QtGui.QTableWidgetItem(msg))
        else:
            self.ui.passwordTable.setItem(row, 4, QtGui.QTableWidgetItem(""))
        self.ui.passwordTable.resizeColumnsToContents()

    def onAddEntry(self):

        if not self.sharedPass:
            self.warnNoConfig()
            return

        dlg = EntryDialog(False, self.sharedPass.listPublicKeys(), self)
        if dlg.exec_() != QtGui.QDialog.Accepted:
            return

        passInfo = dlg.getPassInfo()

        gotWarning = self.addOrUpdateEntry(-1, None, passInfo)
        if gotWarning:
            QtGui.QMessageBox.warning(self, "Bad fingerprints", "Not all specified fingerprints were found in the keys, only used the known ones")

    def appendEntryToTable(self, newEntry):
        row = self.ui.passwordTable.rowCount()
        self.ui.passwordTable.insertRow(row)

        self.setRowEntry(row, newEntry)

    def setRowEntry(self, row, newEntry, message = None):
        # We're going to abuse the first column to attach some extra info

        typeEntry = QtGui.QTableWidgetItem(newEntry[1]["type"])
        typeEntry.entryData = newEntry

        self.ui.passwordTable.setItem(row, 0, typeEntry)
        self.ui.passwordTable.setItem(row, 1, QtGui.QTableWidgetItem(newEntry[1]["description"]))
        self.ui.passwordTable.setItem(row, 2, QtGui.QTableWidgetItem(newEntry[1]["login"]))
        self.ui.passwordTable.setItem(row, 3, QtGui.QTableWidgetItem(newEntry[1]["host"]))
        self.setEntryMessage(row, message)

    def clearStatusColumn(self):
        rows = self.ui.passwordTable.rowCount()
        for r in range(rows):
            self.ui.passwordTable.setItem(r, 4, QtGui.QTableWidgetItem(""))

    def getMultipleSelectedRows(self):
        items = self.ui.passwordTable.selectedItems()

        selRows = set()
        for i in items:
            selRows.add(i.row())

        rows = [ row for row in selRows ]
        rows.sort()
        return rows

    def getSelectedRow(self):
        selRows = self.getMultipleSelectedRows()

        if len(selRows) == 0:
            QtGui.QMessageBox.warning(self, "No items selected", "No items selected")
            return -1

        if len(selRows) != 1:
            QtGui.QMessageBox.warning(self, "Multiple items selected", "Can't work with multiple selected items")
            return -1

        return selRows[0]

    def onEntryDoubleClicked(self, row, col):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        item = self.ui.passwordTable.item(row, 0)
        dlg = EntryDialog(True, self.sharedPass.listPublicKeys(), self)
        dlg.setPassInfo(item.entryData[1])
        dlg.exec_()

    def onCellActivated(self, row, col):
        self.ui.passwordTable.selectRow(row)

    def onCellRightClicked(self, row, col, globalX, globalY):

        item = self.ui.passwordTable.item(row, 0)
        self.popupMenu.process(item.entryData[1], globalX, globalY)

    def filterFingerPrints(self, entryData):
        gotWarning = False
        knownFingerprints = set([ p for (p,n) in self.sharedPass.listPublicKeys() ])
        ownPrint = self.sharedPass.getOwnFingerprint()

        fingerPrints = entryData["fingerprints"]
        filteredPrints = None # 'None' means all fingerprints

        if fingerPrints is not None:
            filteredPrints = set()
            filteredPrints.add(ownPrint) # Always make sure we can decode it ourselves

            for p in fingerPrints:
                if p in knownFingerprints:
                    filteredPrints.add(p)
                else:
                    gotWarning = True

            if not ownPrint in entryData["fingerprints"]:
                entryData["fingerprints"].append(ownPrint)

            tmp = set(entryData["fingerprints"])
            entryData["fingerprints"] = [ p for p in tmp ] # Make sure the entry does not contain any duplicates

        return (gotWarning, filteredPrints) 

    def onChangeEntry(self):
    
        if not self.sharedPass:
            self.warnNoConfig()
            return

        row = self.getSelectedRow()
        if row < 0:
            return

        item = self.ui.passwordTable.item(row, 0)
        dlg = EntryDialog(False, self.sharedPass.listPublicKeys(), self)
        dlg.setPassInfo(item.entryData[1])

        if dlg.exec_() != QtGui.QDialog.Accepted:
            return

        passInfo = dlg.getPassInfo()

        gotWarning = self.addOrUpdateEntry(row, item.entryData[0], passInfo)
        if gotWarning:
            QtGui.QMessageBox.warning(self, "Bad fingerprints", "Not all specified fingerprints were found in the keys, only used the known ones")

    def onRemoveEntry(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        row = self.getSelectedRow()
        if row < 0:
            return
        
        item = self.ui.passwordTable.item(row, 0)
        entryID = item.entryData[0]
        entryDesc = item.entryData[1]["description"]

        warning = "Are you sure you want to remove row %d, for '%s'" % (row+1, entryDesc)
        if QtGui.QMessageBox.question(self, "Sure?", warning, QtGui.QMessageBox.Yes|QtGui.QMessageBox.No, QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
            return

        self.sharedPass.removePassword(entryID) 
        self.ui.passwordTable.removeRow(row)

    def sortPassEntries(self, entries):
        tmp = [ ]
        for e in entries:
            desc = e[1]["description"].lower()
            tmp.append( (desc, e) )

        tmp.sort()
        return [ e for (d,e) in tmp ]

    def onSyncPassInfo(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        sp = self.sharedPass
        #(ret,undec,allfiles) = sp.syncPassInfo()

        (r, msg) = decodingpasswordsdialog.syncPassInfo(sp, self)
        if not r:
            QtGui.QMessageBox.warning(self, "Error trying to decode password entries", msg)
            return

        (ret,undec,allfiles) = r

        allPassInfo = sp.listPasswords()
        allPassInfo = self.sortPassEntries(allPassInfo)

        while self.ui.passwordTable.rowCount():
            self.ui.passwordTable.removeRow(0)

        for entry in allPassInfo:
            self.appendEntryToTable(entry)

        self.statusBar().showMessage("Decoded %d of %d password files" % (len(allfiles)-len(undec),len(allfiles)), 10000)
        if not ret: # There are some entries that could not be decrypted
            if len(undec) == len(allfiles):
                QtGui.QMessageBox.warning(self, "Can't decode any file", "No password files could be decoded.\nAre you allowed to view them?")

    def onSyncPublicKeys(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        self.sharedPass.syncPublicKeys()

    def onRecodeAll(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        warning = "Are you sure you want to re-encode all entries?"
        if QtGui.QMessageBox.question(self, "Sure?", warning, QtGui.QMessageBox.Yes|QtGui.QMessageBox.No, QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
            return

        warning = "Not all fingerprints found for:"
        gotWarning = False

        rows = self.ui.passwordTable.rowCount()
        for r in range(rows):
            item = self.ui.passwordTable.item(r, 0)
            entryID = item.entryData[0]
            entryData = item.entryData[1]

            (gotSubWarning, filteredPrints) = self.filterFingerPrints(entryData)

            if gotSubWarning:
                warning += "\n * " + entryData["description"]
                gotWarning = True

            self.sharedPass.updatePassword(entryID, entryData, filteredPrints)

        if not gotWarning:
            QtGui.QMessageBox.information(self, "Done", "All entries have been re-encoded with specified keys")
        else:
            QtGui.QMessageBox.warning(self, "Can't encode all", warning)

    def warnNoConfig(self):
        QtGui.QMessageBox.warning(self, "No valid config settings processed", "No valid config settings processed")

    def getPassPhrase(self):
        dlg = PassphraseDialog(self)
        if dlg.exec_() != QtGui.QDialog.Accepted or dlg.getPassphrase() == None:
            return False
        
        self.passPhrase = dlg.getPassphrase()
        return True

    def onNewPassPhrase(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        if self.getPassPhrase():
            self.sharedPass.setPassPhrase(self.passPhrase)
            #QtGui.QMessageBox.information(self, "Done", "New passphrase set, reloading password info")
            self.onSyncPassInfo()

    def onConfigure(self):

        dlg = ConfigDialog(self.getConfigDir(), self.confData, self)
        if dlg.exec_() != QtGui.QDialog.Accepted:
            return

        try:
            conf = dlg.getConfigurationData()
            if not os.path.isdir(conf["PubKeyDir"]):
                raise SherPassException("Specified public key path does not seem to exist")
            if not os.path.isdir(conf["PassDir"]):
                raise SherPassException("Specified password info directory does not seem to exist")
    
            if not os.path.exists(self.getConfigDir()):
                os.makedirs(self.getConfigDir(), 0700)

            confFileName = self.getConfigFile()
            if os.path.exists(confFileName):
                if QtGui.QMessageBox.question(self, "Overwrite config file?", "Are you sure you want to overwrite %s" % confFileName, QtGui.QMessageBox.Yes|QtGui.QMessageBox.No, QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
                    return

            newConfData = { "PubKeyDir": conf["PubKeyDir"],
                            "PassDir": conf["PassDir"],
                            "SshCommand": conf["SshCommand"],
                            "TcpPassPhrase": conf["TcpPassPhrase"],
                            "TcpPort": conf["TcpPort"],
                            "AccessUrl": conf["AccessUrl"],
                            "SshPassEnv": conf["SshPassEnv"] }

            if self.confData and self.confData["PrivateKey"]:
                newConfData["PrivateKey"] = self.confData["PrivateKey"]

            if conf["PrivKey"]: # New key specified
                ret = QtGui.QInputDialog.getText(None, "Enter key name", "Enter the base name for the new keys (suffixes will be added)")
                if not ret[1]:
                    return

                keyName = ret[0].strip()
                if not keyName:
                    raise SherPassException("Specified key name cannot be empty")
                if "/" in keyName or "\\" in keyName:
                    raise SherPassException("Invalid character in specified key name '%s'" % keyName)

                privKeyPath = os.path.join(self.getConfigDir(), keyName + ".privkey")
                if os.path.exists(privKeyPath):
                    raise SherPassException("Path '%s' for the private key already exists" % privKeyPath)

                pubKeyPath = os.path.join(conf["PubKeyDir"], keyName + ".pubkey")
                if os.path.exists(pubKeyPath):
                    raise SherPassException("Path '%s' for the public key already exists" % pubKeyPath)

                with open(privKeyPath, "wt") as f:
                    f.write(conf["PrivKey"])
                    f.close()
                os.chmod(privKeyPath, 0600)

                with open(pubKeyPath, "wt") as f:
                    f.write(conf["PubKey"])
                    f.close()
                os.chmod(pubKeyPath, 0600)

                newConfData["PrivateKey"] = privKeyPath

            else:
                if not "PrivateKey" in newConfData:
                    raise SherPassException("No private key exists yet and none was imported or generated")

            dataStr = json.dumps(newConfData)

            with open(confFileName, "wt") as f:
                f.write(dataStr)
                f.close()

            os.chmod(confFileName, 0600)

            QtGui.QMessageBox.information(self, "Done", "Configuration saved, restarting")

            self.restart()

        except Exception as e:
            QtGui.QMessageBox.warning(self, "Error setting new configuration", "Error setting new configuration: " + str(e))

    def onSSH(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        row = self.getSelectedRow()
        if row < 0:
            return

        item = self.ui.passwordTable.item(row, 0)
        entry = item.entryData[1]

        if entry["type"] != "SSH":
            QtGui.QMessageBox.warning(self, "Wrong type", "Selected entry does not have the type 'SSH'")
            return

        host = entry["host"]
        user = entry["login"]
        password = entry["password"]

        port = 22
        idx = host.find(":")
        if idx > 0:
            portStr = host[idx+1:]
            try:
                port = int(portStr)
                host = host[:idx]
            except:
                pass # If the port part is not a number, we'll keep using port 22 and the full hostname

        if not "SshCommand" in self.confData:
            QtGui.QMessageBox.warning(self, "Bad config", "The SSH command to execute was not configured")
            return

        sshCommand = shlex.split(self.confData["SshCommand"])
        substCommand = [ ]
        for i in sshCommand:
            e = i.replace("%h", host)
            e = e.replace("%u", user)
            e = e.replace("%p", password)
            e = e.replace("%P", str(port))
            substCommand.append(e)

        env = copy.deepcopy(os.environ)

        if "SshPassEnv" in self.confData and self.confData["SshPassEnv"]:
            env[self.confData["SshPassEnv"]] = password

        try:
            pid = subprocess.Popen(substCommand, env=env)
        except Exception as e:
            QtGui.QMessageBox.warning(self, "Error launching SSH command", "Can't start SSH:\n" + str(e))

    def startTCPListener(self):
        
        server = QtNetwork.QTcpServer()
        port = self.confData["TcpPort"]
        ret = server.listen(QtNetwork.QHostAddress.LocalHost, port)
        if not ret:
            QtGui.QMessageBox.warning(self, "Can't start TCP server", "Unable to start TCP server on port " + str(port))
            return

        self.connectionLog = ""
        self.appendToConnectionLog("TCP server is listening on port " + str(port))

        self.tcpServer = server
        self.badLocationCount = 0
        server.newConnection.connect(self.onNewConnection) # Can't EW() this since we'll also be disconnecting it

    def appendToConnectionLog(self, msg):
        self.connectionLog += time.strftime("%Y-%m-%d %H:%M:%S %Z: ") + msg + "\n"
        print self.connectionLog
        self.statusBar().showMessage(msg, 10000)

    def onNewConnection(self):

        conn = self.tcpServer.nextPendingConnection()
        conn.headerLines = [ ]
        conn.doneReading = False
        conn.readyRead.connect(self.onReadyRead)
        conn.disconnected.connect(self.onDisconnected) # Also emitted when we close the connection

        self.appendToConnectionLog("New connection from: " + conn.peerAddress().toString() + ":" + str(conn.peerPort()))

        self.connections.append( (conn, time.time()) )

    def onReadyRead(self):

        conn = self.sender()
        while conn.canReadLine():
            l0 = conn.readLine()
            if conn.doneReading:
                continue

            l = ""
            for c in l0:
                l += c

            l = l.rstrip()

            #print "Received:" + l

            if len(conn.headerLines) == 0:
                parts = l.split(" ")   
                if len(parts) != 3 or not parts[0] in [ "GET", "OPTIONS" ]:
                    self.appendToConnectionLog("Invalid header line '%s' from %s:%d" % (l,conn.peerAddress().toString(),conn.peerPort()))
                    conn.doneReading = True 
                    conn.close()
                    return

                conn.requestType = parts[0]
                conn.requestLocation = parts[1]

            if len(l) == 0:
                conn.doneReading = True
                self.analyzeRequest(conn)
            elif len(l) < 1024 and len(conn.headerLines) < 1024:
                conn.headerLines.append(l)
        
    def onDisconnected(self):

        conn = self.sender()

        self.appendToConnectionLog("Closed connection for %s:%d" % (conn.peerAddress().toString(),conn.peerPort()))

        idx = -1
    
        for i in range(len(self.connections)):
            if self.connections[i][0] == conn:
                idx = i
                break

        if idx < 0:
            print "onDisconnected: connection not found"
        else:
            self.connections = self.connections[0:idx] + self.connections[idx+1:]
    
    def onConnectionCheck(self): # We'll do one at a time for simplicity

        t = time.time()

        for (c, t0) in self.connections:
            if t - t0 > 10:
                self.appendToConnectionLog("Timeout for connection %s:%d" % (c.peerAddress().toString(),c.peerPort()))
                c.close()

    def getLoadedEntriesAsJSON(self, pretty=False):

        data = [ self.ui.passwordTable.item(r, 0).entryData[1] for r in range(self.ui.passwordTable.rowCount()) ]
        if pretty:
            body = json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))
        else:
            body = json.dumps(data)

        return body

    def analyzeRequest(self, conn):
        
        headers = { }

        for l in conn.headerLines[1:]:
            idx = l.find(":")
            if idx > 0:
                headers[l[:idx].lower()] = l[idx+1:]
        
        def startResponse(line, headers):
            response = line + "\r\n"
            response += "Access-Control-Allow-Origin: *\r\n"
            x = "access-control-request-headers"

            if x in headers:
                response += "Access-Control-Allow-Headers:" + headers[x]
                response += "\r\n"

            return response

        if conn.requestType == "OPTIONS":
            response = startResponse("HTTP/1.1 200 OK", headers)
            response += "Content-length: 0\r\n\r\n"
        else:

            idx = conn.requestLocation.find("?")
            subLocation = conn.requestLocation[:idx] if idx >= 0 else conn.requestLocation
            
            if subLocation != "/" + self.confData["AccessUrl"]:
                response = startResponse("HTTP/1.1 404 Not found", headers)
                self.badLocationCount += 1
                self.appendToConnectionLog("Connection %s:%d requested invalid location %s" % (conn.peerAddress().toString(),conn.peerPort(),subLocation))
                if self.badLocationCount == 10:
                    self.scheduleAsync(self.onTooManyBadConnections)

            else:
                response = startResponse("HTTP/1.1 200 OK", headers)
                response += "Content-type: image/png\r\n"

                body = self.getLoadedEntriesAsJSON()
                body = aescrypt.encrypt(body, self.confData["TcpPassPhrase"].encode('ascii','ignore'))
                
                width = 1024
                numLines = len(body)/(width*3) + 1
                totalLen = (width*3)*numLines

                restLen = totalLen - len(body)
                body += ' ' * restLen

                # Encode as an image so we can use an Image() in we web page to retrieve
                # this instead of an XHR

                img = QtGui.QImage(width, numLines, QtGui.QImage.Format_RGB32)
                for y in range(numLines):
                    for x in range(width):
                        idx = (x + y*width)*3

                        r = ord(body[idx])
                        g = ord(body[idx+1])
                        b = ord(body[idx+2])

                        img.setPixel(x, y, QtGui.qRgb(r,g,b))

                ba = QtCore.QByteArray()
                buf = QtCore.QBuffer(ba)
                buf.open(QtCore.QIODevice.WriteOnly)
                img.save(buf, "PNG")
                
                response += "Content-length: " + str(len(ba)) + "\r\n\r\n"
                response += ba
                
                self.appendToConnectionLog("Sent password entries to %s:%d" % (conn.peerAddress().toString(),conn.peerPort()))

        conn.write(response)
        conn.flush()
        conn.close()

    def onTooManyBadConnections(self):
        self.closeAllConnections()
        self.appendToConnectionLog("Too many bad connections!")
        QtGui.QMessageBox.warning(self, "Stopping TCP server", "Detected too many incoming connections with an invalid location identifier.\nStopping TCP server.")

    def onQuit(self):
        self.quitting = True
        self.close()

    def onShowFingerprints(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        dlg = FingerprintDialog(self.sharedPass.listPublicKeys(), True, self)
        dlg.exec_()

    def toggleShow(self, status):
        if not self.isVisible():
            self.show()
        
        self.raise_()
        self.activateWindow()

    def onViewPrivateKey(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        dlg = KeyDialog(self.sharedPass.exportPrivateKey(), self)
        dlg.exec_()

    def onExportPublicKey(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        saveFile = QtGui.QFileDialog.getSaveFileName(self, "Specify file to save public key to", dir = None, filter = "Public keys (*.pubkey)")
        if not saveFile or not saveFile[0]:
            return

        fileName = saveFile[0]
        if os.path.exists(fileName):
            QtGui.QMessageBox.warning(self, "File exits", "The specified file already seems to exist, aborting")
            return

        f = None
        try:
            f = open(fileName, "wt")
            f.write(self.sharedPass.exportPublicKey())
        except Exception as e:
            QtGui.QMessageBox.warning(self, "Error writing to file", "Could not write the public key to the specified file:\n" + str(e))
            return
        finally:
            if f: f.close()

        QtGui.QMessageBox.information(self, "Done", "Public key was written to the specified file")

    def onViewPublicKey(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return

        dlg = KeyDialog(self.sharedPass.exportPublicKey(), self)
        dlg.exec_()
        
    def commonChangePasswordChecks(self):
        if not self.sharedPass:
            self.warnNoConfig()
            return None
        
        allRows = self.getMultipleSelectedRows()
        rows = [ ]
        for r in allRows:
            item = self.ui.passwordTable.item(r, 0)
            entry = item.entryData[1]
            if entry["type"] == "SSH":
                rows.append(r)
        
        if len(rows) == 0:
            QtGui.QMessageBox.warning(self, "No SSH entries selected", "Unable to change password: no SSH entries were selected")
            return None

        return rows

    def onChangeToSpecificPassword(self):
        rows = self.commonChangePasswordChecks()
        if not rows:
            return

        dlg = NewPasswordDialog(self)
        if dlg.exec_() != QtGui.QDialog.Accepted:
            return

        newPassword = dlg.getPassword()
        itemsToChange = [ ]
        for r in rows:
            item = self.ui.passwordTable.item(r,0)
            name = item.entryData[0]
            passInfo = item.entryData[1]
            itemsToChange.append( { "row": r, "name": name, "passInfo": passInfo, "newPassword": newPassword, "log": [ ] } )

        self.changeSSHPasswords(itemsToChange)

    def onChangeToRandomPassword(self):
        rows = self.commonChangePasswordChecks()
        if not rows:
            return

        itemsToChange = [ ]
        log = ""
        for r in rows:
            item = self.ui.passwordTable.item(r,0)
            name = item.entryData[0]
            passInfo = item.entryData[1]
            newPassword = randomstring.getRandomString(16)
            itemsToChange.append( { "row": r, "name": name, "passInfo": passInfo, "newPassword": newPassword, "log": [ ] } )
            log += "%s@%s -> '%s'\n" % (passInfo["login"], passInfo["host"], newPassword)

        dlg = LogDialog("Going to try to install the following random passwords\n" + 
                        "for the following hosts. Please temporarily copy/paste them\n" +
                        "somewhere, in case something goes wrong and the password gets\n" +
                        "installed but not saved into the password entry!", self)
        dlg.setLog(log)
        dlg.exec_()

        self.changeSSHPasswords(itemsToChange)

    def changeSSHPasswords(self, itemsAndPasswords):
        if QtGui.QMessageBox.question(self, "Sure?", "Are you really sure you want to set new SSH passwords on all selected hosts?", QtGui.QMessageBox.Yes|QtGui.QMessageBox.No, QtGui.QMessageBox.No) != QtGui.QMessageBox.Yes:
            return

        self.clearStatusColumn()
        dlg = ChangeSSHPasswordsDialog(itemsAndPasswords, self)
        dlg.exec_()

    def onConnectionLog(self):

        dlg = LogDialog(None, self)
        dlg.setLog(self.connectionLog)
        dlg.exec_()

    def onVersion(self):
        versionString = "Revision: unknown\nDate: unknown\n"

        try:
            import version
            versionString = "Revision: %s\nDate: %s\n" % (version.revisionString, version.dateString)
        except:
            pass

        # Abusing the log dialog for this
        dlg = LogDialog(versionString, self, QtGui.QStyle.SP_MessageBoxInformation, "Version", "Info")
        dlg.setLog(licenses.licenses)
        dlg.exec_()

    def onWebSite(self):
        QtGui.QDesktopServices.openUrl(QtCore.QUrl("https://sherpass-lite.appspot.com/about.html"))

    def onExportEntries(self):
        data = self.getLoadedEntriesAsJSON(True)
        
        fileName, filt = QtGui.QFileDialog.getSaveFileName(self, "Choose file to save passwords to", filter="JSON Text Files (*.json)")
        if not fileName:
            return

        print fileName
        try:
            with open(fileName, "wt") as f:
                f.write(data)
        except Exception as e:
            QtGui.QMessageBox.warning(self, "Can't write to specified file", "Error writing to '%s'\n%s" % (fileName, str(e)))

    def keyPressEvent(self, evt):
        self.ui.filterEdit.keyPressEvent(evt)
        self.ui.filterEdit.setFocus()

    def keyReleaseEvent(self, evt):
        self.ui.filterEdit.keyReleaseEvent(evt)

    def onFilterChanged(self, txt):
        if not txt:
            self.ui.filterEdit.hide()
            self.ui.filterLabel.hide()
        else:
            self.ui.filterEdit.show()
            self.ui.filterLabel.show()

def getDefaultConfigDir():
    homeDir = os.path.expanduser("~")
    configDir = os.path.join(homeDir,".sherpass")
    return configDir

def main():

    app = QtGui.QApplication(sys.argv)
    if len(sys.argv) > 2:
        print "The only allowed (optional) parameter is a base configuration directory"
        sys.exit(-1)

    QtCore.QCoreApplication.setOrganizationDomain("sherpass-lite.appspot.com")
    QtCore.QCoreApplication.setOrganizationName("SherPass")
    QtCore.QCoreApplication.setApplicationName("SherPass")

    confDir = getDefaultConfigDir() if len(sys.argv) == 1 else sys.argv[1]

    mainWin = SherPass(confDir)
    mainWin.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()
