#!/usr/bin/env python

import paramiko
import time
import select
import sys

class LineBuffer(object):
    def __init__(self, chan):
        self.chan = chan
        self.history = [ ]
        self.currentLine = ""
        self.fullLog = ""

    def poll(self):
        r, w, e = select.select([self.chan], [], [], 0.100)
        if self.chan in r:
            x = self.chan.recv(1024)
            if len(x) == 0:
                return False
            
            self.processBytes(x)

        return True

    def processBytes(self, x):
        for c in x:
            self.currentLine += c
            self.fullLog += c
            if c == '\n':
                self.history.append(self.currentLine)
                self.currentLine = ""

    def waitFor(self, possibilities, timeout):
        startTime = time.time()
        while time.time() - startTime < timeout:
            self.poll()
            for s in possibilities:
                if s[-1] == '\n':
                    if len(self.history) > 0 and s[:-1] in self.history[-1]:
                        matchLine = self.history[-1]
                        self.clear()
                        return (matchLine, s)
                else:
                    if s in self.currentLine:
                        matchLine = self.currentLine
                        self.clear()
                        return (matchLine, s)

        raise Exception("Timeout while waiting for " + repr(s))

    def clear(self):
        self.history = [ ]
        self.currentLine = ""

    def getFullLog(self):
        return self.fullLog

class SSHPasswordChanger(object):

    def __init__(self, user, host, port, oldPassword, newPassword):
        self.user = user
        self.host = host
        try:
            self.port = int(port)
        except:
            self.port = port
        self.oldPassword = oldPassword
        self.newPassword = newPassword

    def start(self):
        client = paramiko.SSHClient()
        #client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        self.onConnecting()
        client.connect(self.host, self.port, self.user, self.oldPassword, timeout = 15.0, allow_agent = False, look_for_keys = False)
        self.onConnected()
        chan = client.invoke_shell()

        self.onChangingPassword()
        self.changePassword(chan)
        chan.close()
        client.close()

        # Here we're making sure that the old password doesn't work anymore
        self.onVerifyingOldPasswordDisabled()
        try:
            client.connect(self.host, self.port, self.user, self.oldPassword, timeout = 15.0, allow_agent = False, look_for_keys = False)
            raise Exception("Unable to change old password: still works!")            
        except paramiko.ssh_exception.AuthenticationException as e:
            # Ok, this is expected since we used the wrong password
            client.close()

        self.onTestingNewPassword()
        client.connect(self.host, self.port, self.user, self.newPassword, timeout = 15.0, allow_agent = False, look_for_keys = False)
        client.close()

    def changePassword(self, chan):
        chan.settimeout(0.0)
        lb = LineBuffer(chan)

        chan.send("passwd\n")

        # For root, we don't have to type the old password
        oldPassStrings = [ "Old password:", "(current) UNIX password:" ]
        newPassStrings = [ "ew password:", "new UNIX password" ]

        (matchLine, matchString) = lb.waitFor(oldPassStrings + newPassStrings, 10.0)

        if matchString in oldPassStrings:
            chan.send(self.oldPassword + "\n")
            lb.waitFor([ "ew password:", "new UNIX password" ], 5.0)
            chan.send(self.newPassword + "\n")
            lb.waitFor([ "ew password:", "new UNIX password" ], 5.0)
            chan.send(self.newPassword + "\n")
        else: # directly a new password
            chan.send(self.newPassword + "\n")
            lb.waitFor([ "ew password:", "new UNIX password" ], 5.0)
            chan.send(self.newPassword + "\n")

        lb.waitFor([ "password changed\n", "password updated successfully\n" ], 5.0)

        #print lb.getFullLog()

    def onConnecting(self):
        pass

    def onConnected(self):
        pass

    def onChangingPassword(self):
        pass

    def onVerifyingOldPasswordDisabled(self):
        pass

    def onTestingNewPassword(self):
        pass

def main():
    spc = SSHPasswordChanger(*(sys.argv[1:]))
    spc.start()

if __name__ == "__main__":
    main()
